text = """{"title": "NoviSoul
        novissbm@gmail.com", "href": "http://www.youtube.com/signin?authuser=0&next=%2Fwatch%3Fv%3DEaIYRM1yrM4&action_handle_signin=true", "description": ""},
  {"title": "Sign in", "href": "https://accounts.google.com/ServiceLogin?continue=http%3A%2F%2Fwww.youtube.com%2Fsignin%3Faction_handle_signin%3Dtrue%26hl%3Den_GB%26next%3D%252Fwatch%253Fv%253DEaIYRM1yrM4%26nomobiletemp%3D1&uilel=3&service=youtube&passive=true&hl=en_GB", "description": ""},
  {"title": "Sign up", "href": "http://www.youtube.com/signup?next=%2Fwatch%3Fv%3DEaIYRM1yrM4", "description": ""},
  {"title": "9:58


    
 

Physics of Free Energy Deviceby Eugene Jeong        
        
336,881 views", "href": "http://www.youtube.com/watch?v=EB-jWfzkz_E", "description": ""},
  {"title": "4:49


    
 

[www.witts.ws] Self-Running 40kW (40,000 Watt) Fuelless Generator (1 of 3)by wits2014        
        
488,638 views", "href": "http://www.youtube.com/watch?v=LFu-s6ZmGyE", "description": ""},
  {"title": "2:33


    
 

Free Energy - Evidence of military antigravity technologyby DoubleMarkez        
        
390,020 views", "href": "http://www.youtube.com/watch?v=qljY-YfFaPc", "description": ""},
  {"title": "15:01


    
 

APEX 2013   SSBM L10   Shroomed VS CT EMP Mew2Kingby Jason AxelrodRecommended for you", "href": "http://www.youtube.com/watch?v=pc7v49k5FhY", "description": ""},
  {"title": "161
            
            
              videos
            
          
        
      
          
              
          
          
              
          
      
    
      
        
          
Play all
        
      
  
washby dle3276", "href": "http://www.youtube.com/watch?v=AmcSt5hU4qA&list=PL4517CA6C6244A844", "description": ""},
  {"title": "10:31


    
 

Pyramid Magnet - free energy - english subtitleby MrTermsof        
        
616,081 views", "href": "http://www.youtube.com/watch?v=pMbHswNoGWM", "description": ""},
  {"title": "4:11


    
 

My all new newman motor 1.(TheDaftman)by theDaftman        
        
1,147,470 views", "href": "http://www.youtube.com/watch?v=dL4B_DNBtvc", "description": ""},
  {"title": "2:18


    
 

Is there free energy in magnets?by aetherix01        
        
371,642 views", "href": "http://www.youtube.com/watch?v=vrn5B9a8aOk", "description": ""},
  {"title": "3:00


    
 

The Most Dangerous Video On The Internet  - Trevor Paglenby killuminati63        
        
585,755 views", "href": "http://www.youtube.com/watch?v=9xEuhEHDJM8", "description": ""},
  {"title": "2:18


    
 

Free Energy - Magnet Motorby ATBootstrap        
        
358,641 views", "href": "http://www.youtube.com/watch?v=hfkwCE3BeBs", "description": ""},
  {"title": "2:38


    
 

100% free energy generator is easy to buildby LifeHack2012        
        
238,092 views", "href": "http://www.youtube.com/watch?v=GEUyhhMEs7U", "description": ""},
  {"title": "3:41


    
 

5KW free energy –±–µ—Å—Ç–æ–ø–ª–∏–≤–Ω—ã–π –≥–µ–Ω–µ—Ä–∞—Ç–æ—Ä Kapanadze –ö–∞–ø–∞–Ω–∞–¥–∑–µby Alexander Frolov        
        
488,213 views", "href": "http://www.youtube.com/watch?v=uxQ99R4gOWY", "description": ""},""".split('\n')
sources = ' '.join([te for te in text if te])
while True:
    if '  ' in sources:
        sources = sources.replace('  ',' ').replace('\t',' ')
    else:
        break
sources = sources.replace('}, {','},{').replace('},{','},\n{')
input(sources)





# userAgentManager
import random

operating_systems = ['Macintosh','Windows','Linux']
browsers = ['Firefox','Chrome','IceDragon','Waterfox','Gecko','Safari','MetaSr']

def _pick(val, options):
    if not val: return options[0]
    if val in options: return val
    l = val.lower()
    for o in options:
        if l in o.lower():
            return o
    return options[0]

class UserAgentManager:
    def __init__(self, operating_system=None, browser=None, version=None, user_agent=None):
        self.operating_system = _pick(operating_system, operating_systems)
        self.browser = _pick(browser, browsers)
        self.version = version or '42.0'
        self.user_agent = user_agent or self.get_user_agent()
        self.header = {"user-agent": self.user_agent}

    @staticmethod
    def user_agent_db():
        from ..big_user_agent_list import big_user_agent_dict
        return big_user_agent_dict

    def get_user_agent(self):
        ua_db = self.user_agent_db()
        os_db = ua_db.get(self.operating_system) or random.choice(list(ua_db.values()))
        br_db = os_db.get(self.browser) or random.choice(list(os_db.values()))
        if self.version in br_db:
            return br_db[self.version]
        return random.choice(list(br_db.values()))

class UserAgentManagerSingleton:
    _instance = None

    @staticmethod
    def get_instance(**kwargs):
        ua = kwargs.get("user_agent")
        if UserAgentManagerSingleton._instance is None:
            UserAgentManagerSingleton._instance = UserAgentManager(**kwargs)
        else:
            # rebuild if user_agent explicitly changed
            inst = UserAgentManagerSingleton._instance
            if ua and ua != inst.user_agent:
                UserAgentManagerSingleton._instance = UserAgentManager(**kwargs)
        return UserAgentManagerSingleton._instance



from .url_utils import *
def reconstructUrlFromUrlParse(url=None,parsed=None,parsed_dict=None):
    keys = ['scheme','netloc','path','params','query','fragment']
    if url and not parsed:
        parsed_dict = parse_url(url=url)
    if parsed and not parsed_dict:
        parsed_dict = parse_url(parsed=parsed)
    if parsed_dict:
        scheme = parsed_dict.get('scheme')
        nuUrl =''
        for key in keys:
            value = parsed_dict.get(key,'')
            if key  == 'scheme':
                nuUrl += f'{value}://' if value else ''
            elif key  == 'query':
                nuUrl += f'?{reconstructQuery(value)}' if value else ''
            elif key  == 'netloc':
                nuUrl += reconstructNetLoc(value) if value else ''
            else:
                print(value)
                nuUrl += value
        return nuUrl
    return url
def get_youtube_url(url=None,parsed=None, parsed_dict =None ):
    if url and not parsed:
        parsed_dict = parse_url(url=url)
    if parsed and not parsed_dict:
        parsed_dict = parse_url(parsed=parsed)
    if parsed_dict:
        netloc = parsed_dict.get("netloc")
        domain = netloc.get('domain')
        query = parsed_dict.get('query')
        path = parsed_dict.get('path')
        if domain.startswith('youtu'):
            netloc['www']=True
            netloc['domain'] ='youtube'
            netloc['extention'] = '.com'
            parsed_dict['netloc']=netloc
            v_query = query.get('v')
            if path.startswith('/watch'):
                parsed_dict["path"] = f"/{v_query}"
            elif path.startswith('/shorts'):
                parsed_dict["path"] = path[len('/shorts'):]
            parsed_dict["query"] = {"v":eatAll(parsed_dict["path"],['/'])}
            parsed_dict["path"]='/watch'
            return reconstructUrlFromUrlParse(parsed_dict=parsed_dict)
def get_threads_url(url=None,parsed=None, parsed_dict =None ):
    if url and not parsed:
        parsed = parse_url(url)
    if parsed and not parsed_dict:
        parsed_dict = parse_url(parsed=parsed)
    if parsed_dict:
        netloc = parsed_dict.get("netloc")
        domain = netloc.get('domain')
        if domain.startswith('threads'):
            netloc['www']=True
            netloc['domain'] ='youtube'
            netloc['extention'] = '.net'
            parsed['netloc']=netloc
            return reconstructUrlFromUrlParse(url=url,parsed=parsed,parsed_dict=None)  
def get_corrected_url(url=None,parsed=None, parsed_dict =None ):
    if url and not parsed:
        parsed_dict = parse_url(url=url)
    if parsed and not parsed_dict:
        parsed_dict = parse_url(parsed=parsed)
    if parsed_dict:
        funcs = [get_threads_url,get_youtube_url,reconstructUrlFromUrlParse]
        for func in funcs:
            corrected_url = func(url=url,parsed=parsed,parsed_dict=parsed_dict)
            if corrected_url:
                return corrected_url




from .imports import *
_LOCK = threading.RLock()
VIDEO_ENV_KEY = "DATA_DIRECTORY"
# Full schema
VIDEO_SCHEMA = {
    "video_path": "video.mp4",
    "info_path": "info.json",
    "audio_path": "audio.wav",
    "whisper_path": "whisper.json",
    "captions_path": "captions.srt",
    "metadata_path": "metadata.json",
    "thumbnail_path": "thumb.jpg",
    "thumbnails_path": "thumbnails.json",
    "total_info_path": "total_info.json",
    "total_aggregated_path": "total_aggregated.json",
    "aggregated_directory": "aggregated",
    "aggregated_dir": {
        "aggregated_json_path": "aggregated.json",
        "aggregated_metadata_path": "aggregated_metadata.json",
        "best_clip_path": "best_clip.txt",
        "hashtags_path": "hashtags.txt",
    },
    "thumbnails_directory": "thumbnails",
    "thumbnails_dir": {
        "frames": "{video_id}_frame_{i}.jpg",  # pattern
    }
}

def download_image(url, save_path=None):
    """
    Downloads an image from a URL and saves it to the specified path.
    
    Args:
        url (str): The URL of the image to download
        save_path (str, optional): Path to save the image. If None, uses the filename from URL
        
    Returns:
        str: Path where the image was saved, or None if download failed
    """
    try:
        # Send GET request to the URL
        response = requests.get(url, stream=True)
        
        # Check if the request was successful
        if response.status_code == 200:
            # Set decode_content=True to automatically handle Content-Encoding
            response.raw.decode_content = True
            
            # If no save_path provided, extract filename from URL
            if save_path is None:
                # Get filename from URL
                filename = url.split('/')[-1]
                save_path = filename
            
            # Ensure the directory exists
            os.makedirs(os.path.dirname(save_path), exist_ok=True)
            
            # Write the image content to file
            with open(save_path, 'wb') as f:
                f.write(response.content)
            
            print(f"Image successfully downloaded to {save_path}")
            return save_path
        else:
            print(f"Failed to download image. Status code: {response.status_code}")
            return None
            
    except requests.exceptions.RequestException as e:
        print(f"Error downloading image: {str(e)}")
        return None
    except Exception as e:
        print(f"An unexpected error occurred: {str(e)}")
        return None
def get_thumbnails(directory,info):
    thumbnails_dir = os.path.join(directory,'thumbnails')
    os.makedirs(thumbnails_dir, exist_ok=True)
    thumbnails = info.get('thumbnails',[])
    for i,thumbnail_info in enumerate(thumbnails):
        thumbnail_url = thumbnail_info.get('url')
        thumbnail_base_url = thumbnail_url.split('?')[0]
        baseName = os.path.basename(thumbnail_base_url)
        fileName,ext = os.path.splitext(baseName)
        baseName = f"{fileName}{ext}"
        resolution = info['thumbnails'][i].get('resolution')
        if resolution:
            baseName = f"{resolution}_{baseName}"
        img_id = info['thumbnails'][i].get('id')
        if img_id:
            baseName = f"{img_id}_{baseName}"
        thumbnail_path = os.path.join(thumbnails_dir,baseName)
        info['thumbnails'][i]['path']=thumbnail_path
        download_image(thumbnail_url, save_path=thumbnail_path)
    return info

def optimize_video_for_safari(input_file, reencode=False):
    """
    Optimizes an MP4 file for Safari by moving the 'moov' atom to the beginning.
    Optionally, re-encodes the video for maximum compatibility.
    
    Args:
        input_file (str): Path to the original MP4 file.
        reencode (bool): If True, re-encode the video for Safari compatibility.
        
    Returns:
        str: Path to the optimized MP4 file.
    """
    tmp_dir = tempfile.mkdtemp()
    try:
        local_input = os.path.join(tmp_dir, os.path.basename(input_file))
        shutil.copy2(input_file, local_input)
        
        base, ext = os.path.splitext(local_input)
        local_output = f"{base}_optimized{ext}"
        
        if reencode:
            # Re-encoding command for maximum Safari compatibility
            command = [
                "ffmpeg", "-i", local_input,
                "-c:v", "libx264", "-profile:v", "baseline", "-level", "3.0", "-pix_fmt", "yuv420p",
                "-c:a", "aac", "-b:a", "128k",
                "-movflags", "faststart",
                local_output
            ]
        else:
            # Simple faststart with stream copy
            command = [
                "ffmpeg", "-i", local_input,
                "-c", "copy", "-movflags", "faststart",
                local_output
            ]
        
        try:
            subprocess.run(command, check=True)
            shutil.copy2(local_output, input_file)
            print(f"Optimized video saved as {input_file}")
        except subprocess.CalledProcessError as e:
            print(f"Error during optimization: {e}")
        return input_file
    finally:
        shutil.rmtree(tmp_dir)




def dl_video(url, download_directory=None, output_filename=None,
             get_info=None, download_video=None, ydl_opts=None):
    mgr = get_video_info(
        url,
        download_directory=download_directory,
        output_filename=output_filename,
        get_info=get_info,
        download_video=download_video,
        ydl_opts=ydl_opts,  # pass through
    )
    return get_video_info_from_mgr(mgr)
def for_dl_video(url,download_directory=None,output_filename=None,get_info=None,download_video=None):
    get_info = bool_or_default(get_info,default=True)
    download_video =bool_or_default(download_video,default=True)
    video_mgr = dl_video(url,download_directory=download_directory,output_filename=output_filename,get_info=get_info,download_video=download_video)
    if get_video_info_from_mgr(video_mgr):
        return video_mgr
    videos = soupManager(url).soup.find_all('video')
    for video in videos:
        src = video.get("src")
        video_mgr = dl_video(src,download_directory=download_directory,output_filename=output_filename,download_video=download_video)
        if get_video_info_from_mgr(video_mgr):
            return video_mgr
def downloadvideo(url,
                  directory=None,
                  output_filename=None,
                  rename_display=None,
                  thumbnails=None,
                  audio=None,
                  safari_optimize=None,
                  download_video=None,
                  flat_layout: bool = False,
                  *args, **kwargs):
    """
    Download a video and save alongside info.json and other schema files.
    If flat_layout=True, everything goes into `directory` directly.
    """
    rename_display = bool_or_default(rename_display)
    thumbnails = bool_or_default(thumbnails)
    audio = bool_or_default(audio, default=False)
    safari_optimize = bool_or_default(safari_optimize, default=True)
    download_video = bool_or_default(download_video, default=True)

    output_filename = output_filename or get_temp_file_name(url)
    video_mgr = for_dl_video(url,
                             download_directory=directory,
                             output_filename=output_filename,
                             download_video=download_video)
    info = video_mgr.info
    display_id = get_display_id(info)

    os.makedirs(directory, exist_ok=True)
    video_directory = directory if flat_layout else os.path.join(directory, display_id)
    os.makedirs(video_directory, exist_ok=True)
    info['file_path'] = video_directory
    if info:
        file_path = info.get('file_path')
    if rename_display and file_path:
        # Rename using metadata
        video_id = info.get('id', get_temp_id(url))
        title = output_filename or get_video_title(info)
        safe_title = get_safe_title(title)
        final_filename = output_filename or f"{safe_title}_{video_id}"
        final_filename = f"{final_filename}.mp4"
        new_path = os.path.join(video_directory, final_filename)
        if os.path.exists(info['file_path']):
            os.rename(info['file_path'], new_path)
            info['file_path'] = new_path
        info['file_path'] = new_path
            
            # *** Here we call the optimization function ***
    video_path = info.get('file_path')
    if video_path and video_path.lower().endswith('.mp4') and safari_optimize:
        info['file_path'] = optimize_video_for_safari(video_path,reencode=safari_optimize)
    info_path = os.path.join(video_directory, 'info.json')
    if thumbnails:
        info = get_thumbnails(video_directory, info)
    if audio:
        try:
            info = download_audio(directory, info)
        except:
            info['audio_path'] = None
    info['json_path'] = os.path.join(video_directory, 'info.json')
    safe_dump_to_file(info, info['json_path'])

    # attach schema paths with flat awareness
    info = ensure_standard_paths(info, directory, flat_layout=flat_layout)
    return info


def expand_schema(video_id: str, folder: str, schema: dict[str, Any], flat_layout: bool = False) -> dict[str, Any]:
    """
    Expand VIDEO_SCHEMA into concrete paths (recursively).
    - Replaces {video_id} placeholder with the actual ID.
    - If flat_layout=True, do NOT create a <video_id> subdir.
    """
    result = {}
    for key, rel in schema.items():
        if isinstance(rel, dict):
            dirname = key.replace("_dir", "").replace("_directory", "")
            # If flat, stay in current folder
            subfolder = folder if flat_layout else os.path.join(folder, dirname)
            os.makedirs(subfolder, exist_ok=True)
            result[f"{dirname}_directory"] = subfolder
            result[key] = expand_schema(video_id, subfolder, rel, flat_layout=flat_layout)
        elif isinstance(rel, str):
            rel = rel.format(video_id=video_id, i="{i}")
            path = os.path.join(folder, rel)
            if key.endswith("_dir"):
                os.makedirs(path, exist_ok=True)
            result[key] = path
    return result


def ensure_standard_paths(info: dict, video_root: str, flat_layout: bool = False) -> dict:
    """
    Ensure standard paths exist inside <video_root>/ or <video_root>/<video_id>/.
    Controlled by flat_layout.
    """
    vid = info.get("video_id") or info.get("id")
    if not vid:
        return info

    dirbase = video_root if flat_layout else os.path.join(video_root, vid)
    os.makedirs(dirbase, exist_ok=True)
    info["directory"] = dirbase

    schema_paths = expand_schema(vid, dirbase, VIDEO_SCHEMA, flat_layout=flat_layout)

    # flatten for convenience
    def flatten(d, parent_key="", sep="_"):
        flat = {}
        for k, v in d.items():
            new_key = f"{parent_key}{sep}{k}" if parent_key else k
            if isinstance(v, dict):
                flat.update(flatten(v, new_key, sep))
            else:
                flat[new_key] = v
        return flat

    flat_paths = flatten(schema_paths)
    for k, v in flat_paths.items():
        if not info.get(k):
            info[k] = v

    info["schema_paths"] = schema_paths
    return info


def get_video_folder(video_id, envPath=None, flat_layout: bool = False):
    """Return the canonical video folder (with optional flat layout)."""
    root = get_video_directory(envPath=envPath)
    dir_path = root if flat_layout else os.path.join(root, video_id)
    os.makedirs(dir_path, exist_ok=True)
    return dir_path


def get_video_paths(video_id, envPath=None, flat_layout: bool = False):
    """Return dict of canonical paths for this video_id."""
    folder = get_video_folder(video_id, envPath=envPath, flat_layout=flat_layout)
    return expand_schema(video_id, folder, VIDEO_SCHEMA, flat_layout=flat_layout)

def get_video_env(key=None, envPath=None):
    """Pull video directory from env file or environment variables."""
    key = key or VIDEO_ENV_KEY
    return get_env_value(key=key, path=envPath)

def get_video_root(video_root=None):
    """Fallback root directory if no env override is found."""
    home = os.path.expanduser("~")
    candidates = [
        video_root,
        os.path.join(home, "videos"),
        os.path.join(home, "Videos"),
        os.path.join(home, "Downloads"),
        os.path.join(home, "downloads"),
        home,
    ]
    for directory in candidates:
        if directory and os.path.isdir(directory):
            return directory
    return home  # last resort

def get_video_directory(key=None, envPath=None):
    """Assure that a valid video directory exists and return its path."""
    video_directory = get_video_env(key=key, envPath=envPath)
    if not video_directory:
        video_directory = get_video_root()

    os.makedirs(video_directory, exist_ok=True)
    return video_directory

def get_video_folder(video_id, envPath=None):
    """Return the canonical per-video folder and ensure subdirs exist."""
    root = get_video_directory(envPath=envPath)
    dir_path = os.path.join(root, video_id)
    os.makedirs(dir_path, exist_ok=True)

    # Ensure schema directories exist
    for key, rel in VIDEO_SCHEMA.items():
        if rel.endswith("/") or "dir" in key:
            os.makedirs(os.path.join(dir_path, rel), exist_ok=True)

    return dir_path

def get_video_paths(video_id, envPath=None):
    """Return dict of canonical paths for this video_id."""
    folder = get_video_folder(video_id, envPath=envPath)
    return {key: os.path.join(folder, rel) for key, rel in VIDEO_SCHEMA.items()}


def generate_video_id(path: str, max_length: int = 50) -> str:
    # 1. Take basename (no extension)
    file_parts = get_file_parts(path)
    base= file_parts.get("filename")
    if base == 'video':
        base = file_parts.get("dirbase")
    # 2. Normalize Unicode → ASCII
    base = unicodedata.normalize('NFKD', base).encode('ascii', 'ignore').decode('ascii')
    # 3. Lower-case
    base = base.lower()
    # 4. Replace non-alphanumeric with hyphens
    base = re.sub(r'[^a-z0-9]+', '-', base).strip('-')
    # 5. Collapse duplicates
    base = re.sub(r'-{2,}', '-', base)
    # 6. Optionally truncate & hash for uniqueness
    if len(base) > max_length:
        h = hashlib.sha1(base.encode()).hexdigest()[:8]
        base = base[: max_length - len(h) - 1].rstrip('-') + '-' + h
    return base
def _get_video_info(url=None,file_path=None, ydl_opts=None,output_filename=None, cookies_path=None):
    from yt_dlp import YoutubeDL
    
    
    ydl_opts = {
        'quiet': True,
        'skip_download': True,
    }

    if cookies_path and os.path.exists(cookies_path):
        ydl_opts['cookiefile'] = cookies_path

    try:
        with YoutubeDL(ydl_opts) as ydl:
            info = ydl.extract_info(url, download=False)
            return info
    except Exception as e:
        print(f"Failed to extract video info: {e}")
        return None

def get_sha12(s: str) -> str:
    return hashlib.sha1(s.encode("utf-8")).hexdigest()[:12]

def get_atomic_write(path: str, data: dict):
    tmp = f"{path}.tmp"
    safe_dump_to_file(data, tmp)
    os.replace(tmp, path)

def get_normalize_ascii(s: str) -> str:
    return unicodedata.normalize('NFKD', s).encode('ascii', 'ignore').decode('ascii')

def generate_video_id(path_or_title: str, max_length: int = 50) -> str:
    base = os.path.splitext(os.path.basename(path_or_title))[0]
    if base == 'video':
        base = os.path.basename(os.path.dirname(path_or_title)) or base
    base = get_normalize_ascii(base.lower())
    base = re.sub(r'[^a-z0-9]+', '-', base).strip('-')
    base = re.sub(r'-{2,}', '-', base)
    if len(base) > max_length:
        h = get_sha12(base)
        base = f"{base[:max_length - len(h) - 1].rstrip('-')}-{h}"
    return base or get_sha12(path_or_title)

def make_video_info(filepath: str) -> dict:
    import json as _json
    import subprocess as _sub
    cmd = [
        "ffprobe","-v","quiet","-print_format","json",
        "-show_format","-show_streams", filepath
    ]
    probe = _sub.check_output(cmd)
    data = _json.loads(probe)
    info = {
        "id": generate_video_id(filepath),
        "title": os.path.splitext(os.path.basename(filepath))[0],
        "upload_date": datetime.fromtimestamp(os.path.getmtime(filepath)).strftime("%Y%m%d"),
        "duration": float(data["format"].get("duration", 0.0)),
        "streams": data.get("streams", []),
        "format": data.get("format", {}),
        "file_path": os.path.abspath(filepath),
    }
    return info

def get_yt_dlp_info(url: str, ydl_opts: dict | None = None) -> dict | None:
    from yt_dlp import YoutubeDL
    opts = {'quiet': True, 'skip_download': True}
    if ydl_opts:
        opts.update(ydl_opts)
    try:
        with YoutubeDL(opts) as ydl:
            return ydl.extract_info(url, download=False)
    except Exception:
        return None



# --- NEW helpers: unique temp profile + free port + options builder ---
from ..imports import *
def _free_port() -> int:
    s = socket.socket()
    s.bind(("127.0.0.1", 0))
    port = s.getsockname()[1]
    s.close()
    return port

def _make_profile_dir(base="/var/tmp/selenium-profiles") -> str:
    os.makedirs(base, exist_ok=True)
    return tempfile.mkdtemp(prefix="cw-", dir=base)

def _make_chrome_options(binary_path: str | None = None,
                         user_data_dir: str | None = None) -> tuple[Options, str]:
    opts = Options()
    if binary_path:
        opts.binary_location = binary_path
    opts.add_argument("--headless=new")
    opts.add_argument("--no-sandbox")
    opts.add_argument("--disable-dev-shm-usage")
    opts.add_argument("--disable-gpu")
    opts.add_argument("--disable-software-rasterizer")
    opts.add_argument("--disable-extensions")

    prof = user_data_dir or _make_profile_dir()
    opts.add_argument(f"--user-data-dir={prof}")
    opts.add_argument(f"--remote-debugging-port={_free_port()}")

    prefs = {"profile.managed_default_content_settings.images": 2}
    opts.add_experimental_option("prefs", prefs)
    return opts, prof


def _looks_like_html(text_or_bytes: bytes | str) -> bool:
    if not text_or_bytes:
        return False
    s = text_or_bytes if isinstance(text_or_bytes, str) else text_or_bytes.decode("utf-8", "ignore")
    if len(s) < MIN_HTML_BYTES:
        return False
    lowered = s.lower()
    return ("<html" in lowered and "</html>" in lowered) or "<body" in lowered

def _requests_fallback(url: str, headers: dict | None = None, timeout: float = 15.0):
    """Plain requests fallback. Returns `requests.Response | None`."""
    try:
        sess = requests.Session()
        sess.headers.update(headers or {"User-Agent": "Mozilla/5.0"})
        # honor simple redirects and cert issues as needed
        resp = sess.get(url, timeout=timeout, allow_redirects=True, verify=False)
        return resp
    except Exception as e:
        logging.warning(f"requests fallback failed for {url}: {e}")
        return None

def _wait_until_ready(driver, timeout: float = 10.0):
    """Waits for DOM readiness and presence of <body>."""
    try:
        WebDriverWait(driver, timeout).until(
            lambda d: d.execute_script("return document.readyState") in ("interactive", "complete")
        )
    except Exception:
        pass
    try:
        WebDriverWait(driver, timeout).until(EC.presence_of_element_located((By.TAG_NAME, "body")))
    except Exception:
        pass
    # small settle delay for late JS injections
    time.sleep(0.3)
def normalize_url(url, base_url=None):
    manager = seleniumManager(url)
    base_url = manager.base_url
    if url.startswith(base_url):
        url = url[len(base_url):]
    normalized_url = urljoin(base_url, url.split('#')[0])
    if not normalized_url.startswith(base_url):
        return None
    return normalized_url



from ..imports import *
def normalize_ciphers(ciphers: Optional[Union[str, Sequence[str]]]) -> Optional[str]:
    if ciphers is None:
        return None
    if isinstance(ciphers, str):
        # collapse whitespace, dedupe accidental commas
        parts = [p.strip() for p in ciphers.split(",") if p.strip()]
        return ",".join(parts) if parts else ""
    # it's a sequence
    return ",".join(s.strip() for s in ciphers if s and s.strip())



from ..imports import *
def classify_user_agents(user_agents):
    classified = {}

    for ua in user_agents:
        # Detect OS
        os = None
        if "Windows" in ua:
            os = "Windows"
        elif "Macintosh" in ua:
            os = "Macintosh"
        # ... Add other OS checks here

        # Detect browser and version
        browser = None
        version = None
        if "Chrome/" in ua:
            browser = "Chrome"
            version = re.search(r"Chrome/([\d.]+)", ua)
            version = version.group(1) if version else None
        elif "Firefox/" in ua:
            browser = "Firefox"
            version = re.search(r"Firefox/([\d.]+)", ua)
            version = version.group(1) if version else None
        # ... Add other browser checks here

        if os and browser and version:
            if os not in classified:
                classified[os] = {}
            if browser not in classified[os]:
                classified[os][browser] = {}
            classified[os][browser][version] = ua

    return classified
def randomChoice(db):
    if isinstance(db,dict):
        db = list(db.values())
    return random.choice(db)
def getRandomValues(db,key):
    return db.get(key) or randomChoice(db)
def pickUserAgentVars(val, options):
    if not val: return options[0]
    if val in options: return val
    l = val.lower()
    for o in options:
        if l in o.lower():
            return o
    return options[0]



from .imports import *
from .functions import *
class NetworkManager:
    def __init__(self, user_agent_manager=None,user_agent=None, tls_adapter=None,ssl_manager=None, proxies=None, cookies=None,
                 ciphers=None, certification: Optional[str]=None, ssl_options: Optional[List[str]]=None):
        self.ua_mgr = user_agent_manager or UserAgentManager()
        self.ciphers = ciphers or CipherManager().ciphers_string
        self.ssl_mgr = ssl_manager or SSLManager(
            ciphers=self.ciphers or CipherManager().ciphers_string,
            ssl_options=ssl_options,
            certification=certification
        )
        self.certification = self.ssl_mgr.certification
        self.ssl_options = self.ssl_mgr.ssl_options
        self.session = requests.Session()
        self.session.headers.update({
            "User-Agent": user_agent or self.ua_mgr.user_agent,
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8",
            "Accept-Language": "en-US,en;q=0.9",
            "Connection": "keep-alive"
        })
        self.tls_adapter = tls_adapter or TLSAdapter(self.ssl_mgr)
        self.session.mount("https://", self.tls_adapter)
        self.session.mount("http://", HTTPAdapter())

        if proxies:
            self.session.proxies = proxies
        self.proxies = self.session.proxies
        if cookies:
            if isinstance(cookies, requests.cookies.RequestsCookieJar):
                self.session.cookies = cookies
            elif isinstance(cookies, dict):
                jar = requests.cookies.RequestsCookieJar()
                for k,v in cookies.items(): jar.set(k,v)
                self.session.cookies = jar
            # if string: up to you—parse or ignore
        self.cookies = self.session.cookies
        # retries (optional)
        from requests.adapters import Retry
        self.session.adapters['https://'].max_retries = Retry(total=5, backoff_factor=0.5, status_forcelist=[429,500,502,503,504])
        # Normalize cookies on init
        self.cookies = self._normalize_cookies(cookies)
    def _normalize_cookies(self, cookies):
        """Ensure cookies are always a dict of str->str"""
        if not cookies:
            return {}

        # Already a dict
        if isinstance(cookies, dict):
            return {k: str(v) for k, v in cookies.items() if isinstance(v, (str, bytes))}

        # If a RequestsCookieJar, flatten it into a dict
        if isinstance(cookies, requests.cookies.RequestsCookieJar):
            return {c.name: str(c.value) for c in cookies if isinstance(c.value, (str, bytes))}

        logging.warning(f"Dropping invalid cookies object: {cookies!r}")
        return {}
    def init_poolmanager(self, *args, **kwargs):
        kwargs["ssl_context"] = self._ssl_context
        return super().init_poolmanager(*args, **kwargs)



from .src import *
def get_url_mgr(url=None,url_mgr=None):
    return url_mgr or urlManager(url)
def get_url(url=None,url_mgr=None):
    url_mgr = get_url_mgr(url=url,url_mgr=url_mgr)
    return url_mgr.url

